package mhtree;

import cz.muni.fi.disa.similarityoperators.cover.AbstractRepresentation.PrecomputedDistances;
import cz.muni.fi.disa.similarityoperators.cover.HullOptimizedRepresentationV3;
import messif.buckets.BucketStorageException;
import messif.objects.LocalAbstractObject;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

public abstract class Node implements Serializable {

    /**
     * Serialization ID
     */
    private static final long serialVersionUID = 420L;
    protected final InsertType insertType;
    private final ObjectToNodeDistance objectToNodeDistance;
    protected HullOptimizedRepresentationV3 hull;
    protected Node parent;

    Node(PrecomputedDistances distances, InsertType insertType, ObjectToNodeDistance objectToNodeDistance) {
        this.hull = new HullOptimizedRepresentationV3(distances);
        this.hull.build();
        this.insertType = insertType;
        this.objectToNodeDistance = objectToNodeDistance;
    }

    public static InternalNode createParent(List<Node> nodes, PrecomputedDistances distances, InsertType insertType, ObjectToNodeDistance objectToNodeDistance) {
        List<LocalAbstractObject> objects = nodes.stream()
                .map(Node::getObjects)
                .flatMap(Collection::stream)
                .collect(Collectors.toList());

        return new InternalNode(distances.getSubset(objects), insertType, objectToNodeDistance);
    }

    public double getDistance(LocalAbstractObject object) {
        return objectToNodeDistance.getDistance(object, this);
    }

    public double getNearestDistance(LocalAbstractObject object) {
        return ObjectToNodeDistance.NEAREST_HULL_OBJECT.getDistance(object, this);
    }

    public boolean isCovered(LocalAbstractObject object) {
        return hull.isExternalCovered(object);
    }

    public boolean isLeaf() {
        return (this instanceof LeafNode);
    }

    public void setParent(Node parent) {
        this.parent = parent;
    }

    public List<LocalAbstractObject> getHullObjects() {
        return hull.getHull();
    }

    public int getLevel() {
        return parent == null ? 1 : parent.getLevel() + 1;
    }

    @Override
    public String toString() {
        return "Node{hull=" + hull + '}';
    }

    public abstract void addObject(LocalAbstractObject object) throws BucketStorageException;

    public abstract Set<LocalAbstractObject> getObjects();

    public abstract int getHeight();

    public abstract List<Node> getNodesOnLevel(int level);

    protected void rebuildHull(LocalAbstractObject object) {
        List<LocalAbstractObject> objects = new ArrayList<>(getObjects());
        objects.add(object);

        hull = new HullOptimizedRepresentationV3(objects);
        hull.build();
    }

    protected void addNewObject(LocalAbstractObject object) {
        if (isCovered(object)) return;

        if (insertType == InsertType.INCREMENTAL) {
            hull.addHullObject(object);
            return;
        }

        rebuildHull(object);
    }
}
